import greenfoot.*;

/**
 * Class Snake:
 * the head and the body segments are both of this class
 */
public class Snake extends Actor
{
    static Snake head; // holds a reference to the head of the snake
    static Snake tail; // holds a reference to the last body segment of the snake
    static int direction; // the direction the head of the snake is to go
                       // 0 = right, 1 = down, 2 = left, and 3 = up
    
    private Snake follower = null; // holds a reference to the body segment behind this one, if there is one
    
    
    /**
     * Snake Constructor:
     * if first segment created, make it the head and change its image,
     * else tell tail of appendage and reset tail to this new segment
     */
    public Snake()
    {
        // if first snake segment create then make it the head; make it the last body segment, regardless
        if (head == null)
        {
            head = this;
            tail = this;
            setImage(new GreenfootImage("smiley1.png"));
        }
        else
        {
            tail.follower = this;
            tail = this;
        }
    }
    
    /**
     * Method act:
     * only if acting on the head of the snake, moves the snake and controls the growth of the snake
     */
    public void act()
    {
        if (!this.equals(head)) return;
        setDirection(); 
        if (!canMove()) gameOver(); else moveAndGrow();
    }
    
    /**
     * Method moveAndGrow:
     * moves the whole snake; plus, adds a body segment, if chance has it
     */
    private void moveAndGrow()
    { 
        int atX = head.getX(), atY = head.getY(); // save current x and y of head segment as last location
        Snake segment = head; // initialize a variable to hold the current segment (head)
        moveSnakeHead(); // moves only the head
        // the following is 'follow the leader' to move the body segments
        while(segment.follower != null)
        {
            segment = segment.follower; // gets next segment to be the current segment
            int newX = segment.getX(), newY = segment.getY(); // save the current x and y of segment as new location
            segment.setLocation(atX, atY); // moves segment to last segments old location
            atX = newX;
            atY = newY; // make new location to be the last location
        }
        // the following statement uses random chance to add a new body segment
        if (Greenfoot.getRandomNumber(100) < 10) getWorld().addObject(new Snake(), atX, atY);
    }
    
    /**
     * Method gameOver:
     * changes the image of the snake's head to a frowny-face and stops the execution of the game
     */
    private void gameOver()
    {
        head.setImage(new GreenfootImage("smiley3.png"));
        Greenfoot.stop();
    }

    /**
     * Method moveSnakeHead:
     * moves the head of the snake in 'direction'
     */
    private void moveSnakeHead()
    {
        int dx = getXOffset(), dy = getYOffset();
        setLocation(getX()+dx, getY()+dy);
    }        
    
    /**
     * Method canMove
     *
     * @return The return value informs the calling segment whether the head can move
     * in the direction last requested (checks for edge of world and collision with body segment)
     */
    private boolean canMove()
    {
        if (direction == 0 && getX() == getWorld().getWidth()-1) return false;
        if (direction == 1 && getY() == getWorld().getHeight()-1) return false;
        if (direction == 2 && getX() == 0) return false;
        if (direction == 3 && getY() == 0) return false;
        int dx = getXOffset(), dy = getYOffset();
        return getWorld().getObjectsAt(getX()+dx, getY()+dy, null).isEmpty();
    }
    
    /**
     * Method getXOffset:
     * computes the x offset by evaluating the 'direction' variable
     * 
     * @return The return value is the x offset 
     */
    private int getXOffset()
    {
        return ((3-direction)%2)*(1-direction);
    }
    
    /**
     * Method getYOffset:
     * computes the y offset by evaluating the 'direction' variable
     * 
     * @return The return value is the y offset
     */
    private int getYOffset()
    {
        return ((4-direction)%2)*(2-direction);
    }
        
    /**
     * Method setDirection:
     * gets user input, if any, and sets or retains the current direction of the head of the snake;
     * pressing opposing arrows nullifies the change in that direction
     */
    private void setDirection()
    {
        int dx = 0, dy = 0;
        if (Greenfoot.isKeyDown("right")) dx++;
        if (Greenfoot.isKeyDown("left")) dx--;
        if (Greenfoot.isKeyDown("down")) dy++;
        if (Greenfoot.isKeyDown("up")) dy--;
        if (dx*dx == dy*dy) return;
        direction = getDirection(dx, dy);
        setRotation(direction*90);
    }
    
    /**
     * Method getDirection:
     * computes the direction using the x and y offsets
     * 
     * @param dx A parameter representing the x offset
     * @param dy A parameter representing the y offset
     * @return The return value is the direction
     */
    private int getDirection(int dx, int dy)
    {
        return dx*dx*(1-dx)+dy*dy*(2-dy);
    }
}